;-----------------------------------------------------------------------------
; RetroForth for Linux       | This port was originally written by:
; http://www.retroforth.org  | * Tom Novelli *
;-----------------------------------------------------------------------------
include "macros.fasm"
stdin  = 0
stdout = 1
stderr = 2

sys_exit   =     0x01
sys_read   =     0x03
sys_write  =     0x04
sys_open   =     0x05
sys_close  =     0x06
sys_mprotect =   0x4a

macro system A {
    push 0x0
    mov eax, A
    int 0x80
    add esp, 4
}

macro sys.exit A {
    push dword A
    system sys_exit
}

format elf executable
entry _start

_start:
	call init_retroforth
	upop ebx
	jmp ebx
;---------------------------------------------------------------------
code 'emit', emit
        push eax, ebx, ecx, edx
        mov [buffer], eax       ; Put top of datastack in buffer
        push 0x01               ; Number of characters to write
        push buffer             ; The address of the buffer
        push stdout             ; Write to stdout
        system sys_write
        add esp, 3*4
        pop eax, ebx, ecx, edx
        drop
        ret
code 'key', key
        dup
        push ebx, ecx, edx
        push 0x01               ; Number of characters to read
        push buffer             ; The address of the buffer
        push stdin              ; Read from stdin
        system sys_read
        add esp, 3*4
        pop ebx, ecx, edx
        mov eax, [buffer]       ; Put read character on top of dstack
        ret
code 'type',type                ;
        push ebx                ;
        push ecx                ;
        upop ecx                ;
        upop ebx                ;
        or ecx,ecx              ; Is the count zero?
        jz .done                ; If so, we can exit
.a:     push ecx                ; store the count
        dup                     ; Duplicate it
        xor eax,eax             ; Filter out junk bits
        mov al,[ebx]            ; Obtain the byte to show
        inc ebx                 ; increment to the next byte
        call emit               ; Show the character
        pop ecx                 ; Restore the count
        loop .a                 ; Loop, decrementing the count
.done:  pop ecx                 ;
        pop ebx                 ;
next
code 'bye',freebsd_bye
        sys.exit 0x00

buffer dd 0
;---------------------------------------------------------------------
; FreeBSD syscall interface
; Description:
; syscall ( args argcount syscall# -- result )
; args: as few as 0 or as many as 6 arguments
; argcount: the number of arguments
; syscall-number: a number from /usr/include/asm/unistd.h
; This calling convention was borrowed from Marcel Hendrix's port of 
; eforth for Linux.

code 'syscall', freebsd_syscall
        upop edx                ; Save the system call #
.syscall_1:
        cmp eax, 1              ; Is this a one argument system call?
        jnz .syscall_2          ;
        drop                    ; Forget the count
        push esi                ; Save ESI
        push eax                ; Argument #1
        system edx              ; Do the system call (via macro)
        mov ebx, eax            ; Temporarily save result
        add esp, 4              ; Clean up return stack
        pop esi                 ; Restore ESI
        mov eax, ebx            ; Put the result in EAX
	add esi, 4
        jmp .syscall_exit       ;
.syscall_2:                     ; Is this a two argument system call?
        cmp eax, 2              ;
        jnz .syscall_3          ;
        drop                    ;
        push esi                ; Preserve ESI
        push eax                ; Argument #1
        drop                    ;
        push eax                ; Argument #2
        system edx              ; Do the system call
        mov ebx, eax            ; Save the result
        add esp, 2*4            ;
        pop esi                 ; Restore ESI
        mov eax, ebx            ; Put the result on TOS
	add esi, 4
        jmp .syscall_exit       ;
.syscall_3:                     ; Is this a three argument system call?
        cmp eax, 3              ;
        jnz .syscall_4          ;
        drop                    ;
        push esi                ;
        push eax                ; Argument #1
        drop                    ;
        push eax                ; Argument #2
        drop                    ;
        push eax                ; Argument #3
        system edx              ; Do the system call
        mov ebx, eax            ;
        add esp, 3*4            ;
        pop esi                 ;
        mov eax, ebx            ;
	add esi, 4
        jmp .syscall_exit       ;
.syscall_4:                     ;
        cmp eax, 4              ; Is this a four argument system call?
        jnz .syscall_5          ;
        drop                    ;
        push esi                ;
        push eax                ; Argument #1
        drop                    ;
        push eax                ; Argument #2
        drop                    ;
        push eax                ; Argument #3
        drop                    ;
        push eax                ; Argument #4
        system edx              ; Do the system call
        mov ebx, eax            ;
        add esp, 4*4            ;
        pop esi                 ;
        mov eax, ebx            ;
	add esi, 4
        jmp .syscall_exit       ;
.syscall_5:                     ; Five arguments?
        cmp eax, 5              ;
        jnz .syscall_6          ;
        drop                    ;
        push esi                ;
        push eax                ; Argument #1
        drop                    ;
        push eax                ; Argument #2
        drop                    ;
        push eax                ; Argument #3
        drop                    ;
        push eax                ; Argument #4
        drop                    ;
        push eax                ; Argument #5
        system edx              ; Do the system call
        mov ebx, eax            ;
        add esp, 5*4            ;
        pop esi                 ;
        dup                     ;
        mov eax, ebx            ;
	add esi, 4
        jmp .syscall_exit       ;
.syscall_6:                     ; Six arguments?
        cmp eax, 6              ;
        jnz .syscall_exit       ; If not between 1-6, exit
        drop                    ;
        push esi                ;
        push eax                ; Argument #1
        drop                    ;
        push eax                ; Argument #2
        drop                    ;
        push eax                ; Argument #3
        drop                    ;
        push eax                ; Argument #4
        drop                    ;
        push eax                ; Argument #5
        drop                    ;
        push eax                ; Argument #6
        system edx              ; Do the system call
        mov ebx, eax            ;
        add esp, 6*4            ;
        pop esi                 ;
        dup                     ;
        mov eax, ebx            ;
	add esi, 4
.syscall_exit:
	next

; syscall interface ends here
;---------------------------------------------------------------------
include "core.asm"
